/*
 * kernel/power/suspend2_common.h
 *
 * Copyright (C) 2005-2006 Nigel Cunningham <nigel@suspend2.net>
 *
 * This file is released under the GPLv2.
 *
 * Routines for talking to a userspace program that manages storage.
 *
 * The kernel side:
 * - starts the userspace program;
 * - sends messages telling it when to open and close the connection;
 * - tells it when to quit;
 *
 * The user space side:
 * - passes messages regarding status;
 *
 */

#ifdef CONFIG_PM_DEBUG
#define set_debug_state(bit) (test_and_set_bit(bit, &suspend_debug_state))
#define clear_debug_state(bit) (test_and_clear_bit(bit, &suspend_debug_state))
#else
#define set_debug_state(bit) (0)
#define clear_debug_state(bit) (0)
#endif

#define set_result_state(bit) (test_and_set_bit(bit, &suspend_result))
#define clear_result_state(bit) (test_and_clear_bit(bit, &suspend_result))

enum {
	SUSPEND_ABORT_REQUESTED = 1,
	SUSPEND_NOSTORAGE_AVAILABLE,
	SUSPEND_INSUFFICIENT_STORAGE,
	SUSPEND_FREEZING_FAILED,
	SUSPEND_UNEXPECTED_ALLOC,
	SUSPEND_KEPT_IMAGE,
	SUSPEND_WOULD_EAT_MEMORY,
	SUSPEND_UNABLE_TO_FREE_ENOUGH_MEMORY,
	SUSPEND_ENCRYPTION_SETUP_FAILED
};

extern unsigned int nr_suspends;
